@@ -52,6 +52,9 @@
 #include "libgksu.h"
 #include "../libgksuui/gksuui-dialog.h"
 
+static void
+gksu_context_launch_complete (GksuContext *context);
+
 GType
 gksu_error_get_type (void)
 {
@@ -648,7 +651,7 @@ get_gnome_keyring_password (GksuContext *context)
   gnome_keyring_attribute_list_free (attributes);
   if (
       (result == GNOME_KEYRING_RESULT_OK) &&
-      (g_list_length(list) == 1)
+      (g_list_length(list) >= 1)
       )
     {
       GnomeKeyringFound *found = list->data;
@@ -689,6 +692,7 @@ set_gnome_keyring_password (GksuContext *context, gchar *password)
       static GMainLoop *keyring_loop = NULL;
       GnomeKeyringAttributeList *attributes;
       GnomeKeyringAttribute attribute;
+      GnomeKeyringResult result;
 
       gchar *keyring_name;
       gchar *key_name;
@@ -718,6 +722,14 @@ set_gnome_keyring_password (GksuContext *context, gchar *password)
       keyring_name = gconf_client_get_string (gconf_client, BASE_PATH"save-keyring", NULL);
       if (keyring_name == NULL)
 	keyring_name = g_strdup ("session");
+
+      /* make sure the keyring exists; if an error occurs, use
+         the session keyring */
+      result = gnome_keyring_create_sync(keyring_name, NULL);
+      if ((result != GNOME_KEYRING_RESULT_OK) &&
+	  (result != GNOME_KEYRING_RESULT_ALREADY_EXISTS))
+	keyring_name = g_strdup ("session");
+
       gnome_keyring_item_create (keyring_name,
 				 GNOME_KEYRING_ITEM_GENERIC_SECRET,
 				 key_name,
@@ -906,6 +918,8 @@ su_ask_password (GksuContext *context, gchar *prompt,
 	  g_set_error (error, gksu_quark,
 		       GKSU_ERROR_CANCELED,
 		       _("Password prompt canceled."));
+	  if (context->sn_context)
+	    gksu_context_launch_complete (context);
 	}
 
       gtk_widget_destroy (dialog);
@@ -1877,7 +1891,7 @@ gksu_su_full (GksuContext *context,
   if (context->sn_context)
     gksu_context_launch_initiate (context);
 
-  pid = forkpty (&fdpty, NULL, NULL, NULL);
+  pid = forkpty(&fdpty, NULL, NULL, NULL);
   if (pid == 0)
     {
       gchar **cmd = g_malloc (sizeof(gchar*)*7);
@@ -1968,7 +1982,7 @@ gksu_su_full (GksuContext *context,
 	  tcgetattr (fdpty, &tio);
 	  for (count = 0; (tio.c_lflag & ECHO) && count < 15; count++)
 	    {
-	      usleep (30);
+	      usleep (1000);
 	      tcgetattr (fdpty, &tio);
 	    }
 
@@ -1983,14 +1997,22 @@ gksu_su_full (GksuContext *context,
 
 	      /* try to get the password from the GNOME Keyring first */
 	      password = get_gnome_keyring_password (context);
-	      if (context->debug)
-		fprintf (stderr, "password: %s\n", password);
 	      if (password == NULL)
-		password = ask_pass (context, buf, ask_pass_data, error);
+		{
+		  password = ask_pass (context, buf, ask_pass_data, error);
+		  if (context->debug)
+		    {
+		      fprintf (stderr, "no password on keyring\n");
+		      if (password == NULL)
+			fprintf (stderr, "no password from ask_pass!\n");
+		    }
+		}
 	      else
-		used_gnome_keyring = TRUE;
-	      if (context->debug)
-		fprintf (stderr, "password: %s\n", password);
+		{
+		  if (context->debug)
+		    fprintf (stderr, "password from keyring found\n");
+		  used_gnome_keyring = TRUE;
+		}
 	      if (password == NULL || (error && (*error)))
 		{
 		  if (context->debug)
@@ -2168,76 +2190,51 @@ gksu_su_full (GksuContext *context,
 	    }
 	}
 
-      FD_ZERO(&rfds);
-      FD_SET(fdpty, &rfds);
-
-      tv.tv_sec = 0;
-      tv.tv_usec = 100;
-
-      while (!waitpid (pid, &status, WNOHANG))
+      cmdline = g_strdup("bin/su");
+      /* wait for the child process to end or become something other
+	 than su */
+      while ((!waitpid (pid, &status, WNOHANG)) &&
+	     (g_str_has_suffix(cmdline, "bin/su")))
 	{
-	  bzero (buf, 256);
-	  usleep (200);
-
 	  if (cmdline)
 	    g_free (cmdline);
 	  cmdline = get_process_name (pid);
-
-	  usleep(1000);
-
-	  if (select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
-	    {
-	      g_set_error (error, gksu_quark, GKSU_ERROR_PIPEREAD,
-			   _("Could not read from the pipe "
-			     "with the child: %s"),
-			   strerror(errno));
-	      return FALSE;
-	    }
-
-	  if(FD_ISSET(fdpty, &rfds))
-	    {
-	      bzero(buf, 256);
-	      read (fdpty, buf, 255);
-	      fprintf (stderr, "%s", buf);
-	    }
-	  else
-	    FD_SET(fdpty, &rfds);
-	}
-
-      while (1)
-	{
-	  bzero(buf, 256);
-	  if(read (fdpty, buf, 255) <= 0)
-	    break;
-	  fprintf (stderr, "%s", buf);
+	  usleep(100000);
 	}
 
       if (context->sn_context)
 	gksu_context_launch_complete (context);
 
-      if (WIFEXITED(status))
+      bzero(buf, 256);
+      while (read (fdpty, buf, 255) > 0)
 	{
-	  if (WEXITSTATUS(status))
-	    {
-	      if(cmdline)
-		{
-		  /* su already exec()ed something else, don't report
-		   * exit status errors in that case
-		   */
-		  if (!g_str_has_suffix (cmdline, "su"))
-		    {
-		      g_free (cmdline);
-		      return FALSE;
-		    }
-		  g_free (cmdline);
-		}
+	  fprintf (stderr, "%s", buf);
+	  bzero(buf, 256);
+	}
 
-	      if (error == NULL)
-		g_set_error (error, gksu_quark,
-			     GKSU_ERROR_CHILDFAILED,
-			     _("su terminated with %d status"),
-			     WEXITSTATUS(status));
+      if (!WIFEXITED(status))
+	waitpid(pid, &status, 0);
+
+      if (WEXITSTATUS(status))
+	{
+	  if(cmdline)
+	    {
+	      /* su already exec()ed something else, don't report
+	       * exit status errors in that case
+	       */
+	      if (!g_str_has_suffix (cmdline, "su"))
+		{
+		  g_free (cmdline);
+		  return FALSE;
+		}
+	      g_free (cmdline);
 	    }
+
+	  if (error == NULL)
+	    g_set_error (error, gksu_quark,
+			 GKSU_ERROR_CHILDFAILED,
+			 _("su terminated with %d status"),
+			 WEXITSTATUS(status));
 	}
     }
 
@@ -2471,7 +2468,7 @@ gksu_sudo_full (GksuContext *context,
 	  }
 	else
 	  {
-	    if (context->command[i] == '\\')
+	    if (context->command[i] == '\\' && context->command[i+1] != '\\')
 	      i = i + 1;
 	    tmp_arg = g_realloc (tmp_arg, sizeof(gchar)*(j+1));
 	    tmp_arg[j] = context->command[i];
@@ -2619,7 +2616,7 @@ gksu_sudo_full (GksuContext *context,
 	      return FALSE;
 	    }
 
-	  usleep (100);
+	  usleep (1000);
 
 	  fprintf (outfile, "%s\n", password);
 	  fclose (outfile);
@@ -2633,8 +2630,6 @@ gksu_sudo_full (GksuContext *context,
 	  fgets (buffer, 255, infile);
 	  /* this is the status we are interessted in */
 	  fgets (buffer, 255, infile);
-	  /* turn it ON again */
-	  fcntl(parent_pipe[0], F_SETFL, fcntl(parent_pipe[0], F_GETFL) | O_NONBLOCK);
 	}
       else
 	{
@@ -2683,57 +2678,51 @@ gksu_sudo_full (GksuContext *context,
 			       "the system administrator."));
 	    }
 	}
-      while (!waitpid (pid, &status, WNOHANG))
+
+      cmdline = g_strdup("sudo");
+      /* wait for the child process to end or become something other
+	 then sudo */
+      while ((!waitpid (pid, &status, WNOHANG)) &&
+	     (g_str_has_suffix(cmdline, "sudo")))
 	{
 	  if (cmdline)
 	    g_free (cmdline);
 	  cmdline = get_process_name (pid);
-
-	  usleep(1000);
-
-	  bzero(buffer, 256);
-	  if(!fgets (buffer, 255, infile))
-	    continue;
-	  fprintf (stderr, "%s", buffer);
-	}
-
-      /* make sure we did read everything */
-      while (1)
-	{
-	  bzero(buffer, 256);
-	  if(!fread (buffer, sizeof(gchar), 255, infile))
-	    break;
-	  fprintf (stderr, "%s", buffer);
-	  fflush (stderr);
+	  usleep(100000);
 	}
 
       if (context->sn_context)
 	gksu_context_launch_complete (context);
+
+      while (read (parent_pipe[0], buffer, 255) > 0)
+	{
+	  fprintf (stderr, "%s", buffer);
+	  bzero(buffer, 256);
+	}
+
+      if (!WIFEXITED(status))
+	waitpid(pid, &status, 0);
       sudo_reset_xauth (context, xauth, xauth_env);
 
-      if (WIFEXITED(status))
+      if (WEXITSTATUS(status))
 	{
-	  if (WEXITSTATUS(status))
+	  if(cmdline)
 	    {
-	      if(cmdline)
+	      /* sudo already exec()ed something else, don't report
+	       * exit status errors in that case
+	       */
+	      if (!g_str_has_suffix (cmdline, "sudo"))
 		{
-		  /* sudo already exec()ed something else, don't report
-		   * exit status errors in that case
-		   */
-		  if (!g_str_has_suffix (cmdline, "sudo"))
-		    {
-		      g_free (cmdline);
-		      return FALSE;
-		    }
 		  g_free (cmdline);
+		  return FALSE;
 		}
-
-	      if (error == NULL)
-		g_set_error (error, gksu_quark,
-			     GKSU_ERROR_CHILDFAILED,
-			     _("sudo terminated with %d status"),
-			     WEXITSTATUS(status));
+	      g_free (cmdline);
 	    }
+	  if (error == NULL)
+	    g_set_error (error, gksu_quark,
+			 GKSU_ERROR_CHILDFAILED,
+			 _("sudo terminated with %d status"),
+			 WEXITSTATUS(status));
 	}
     }
 
