@@ -33,7 +33,8 @@
 
 G_BEGIN_DECLS
 
-typedef struct
+typedef struct _GksuContext GksuContext;
+struct _GksuContext
 {
   /*
    * Protected
@@ -63,9 +64,22 @@ typedef struct
 
   /* startup notification */
   SnLauncherContext *sn_context;
+  gchar *sn_id;
+  
+  /* ref counting */
+  gint ref_count;
 
   gboolean debug;
-} GksuContext;
+};
+
+#define GKSU_TYPE_CONTEXT gksu_context_get_type()
+GType                gksu_context_get_type  (void);
+GksuContext*         gksu_context_new       (void);
+GksuContext*         gksu_context_ref       (GksuContext *context);
+void                 gksu_context_unref     (GksuContext *context);
+
+GType gksu_error_get_type (void);
+#define GKSU_TYPE_ERROR (gksu_error_get_type ())
 
 typedef enum
 {
@@ -93,9 +107,6 @@ typedef
 void
 (*GksuPassNotNeededFunc) (GksuContext *context, gpointer user_data);
 
-GksuContext*
-gksu_context_new ();
-
 /*
    getters and setters for the configuration
    options
